/*
   Kalender version 2.0  Copyright (C) 1998-2004 Frank Reglin

   Kalender is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   Kalender is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Kalender; see the file COPYING.
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "kalender.h"


/* constants ********** */
#define PAGELINES 65    /* Anzahl der Zeilen pro Seite */
#define WEEKLINES  7    /* Anzahl der Zeilen pro Woche */

#define DEFAULTYEAR "2005"
#define LPTPORT     "LPT1"

/* Druckerauswahl: steuert die nachfolgenden Default-Steuerfolgen */
/* nur ein LPT_ definieren ! */
#undef  LPT_EPSON_ESCP
#define LPT_HP_LJPCL

#ifdef LPT_EPSON_ESCP   /* epson stylus 800 esc/p */
#define LPT_PROLOG "1B 40 1B 55 01 1B 6B 02 1B 78"
#define LPT_NORMAL "1B 35 1B 46"
#define LPT_HILITE "1B 34 1B 45"
#endif

#ifdef LPT_HP_LJPCL     /* hp laserjet 4 pcl */
                 /* Schrift PC8    10 Zeich/Zoll  71 Zeilen/Seite   A4                3 Leerzeil oben 66 Zeilen/Seite  Einseitig   */
#define LPT_PROLOG "1B 28 31 30 55 1B 26 6B 30 53 1B 26 6C 37 31 50 1B 26 6C 32 36 41 1B 26 6C 33 45 1B 26 6C 36 36 46 1B 26 6C 30 53"
#define LPT_NORMAL "1B 28 73 30 42 1B 28 73 30 53"
#define LPT_HILITE "1B 28 73 35 42 1B 28 73 31 53"
#endif


/* types ********** */
typedef struct
{
   int month;        /* 1 = Jan, ... 12 = Dez                  */
   int date;         /* Tag im Monat, 1..31                    */
   int dayofweek;    /* Wochentag, 0=Mo, 1=Di, ... 6=So        */
   int week;         /* Wochennummer, immer Mo bis So dieselbe */
   char holiday;     /* Flag fuer So und Feiertage             */
} DAY;

/* global data ********** */
int year;
char domonth[13];    /* flags, 1..12 used */

FILE * outf;
char outname[250];
int prolog_len;
int normal_len;
int hilite_len;
unsigned char prolog[70];
unsigned char normal[20];
unsigned char hilite[20];

DAY day[367];     /* Der Jahreskalender */
int first[14];    /* index in day[] fuer den ersten Tag des Monats */
int last[14];     /* index in day[] fuer den letzten Tag des Monats */

char * dayname[7] = { "Mo", "Di", "Mi", "Do", "Fr", "Sa", "So" };
char * monname[13] = { 0, "Januar", "Februar", "Mrz", "April", "Mai",
                          "Juni", "Juli", "August", "September", 
                          "Oktober", "November", "Dezember" };

/* proto types: ********** */
int mprintf( char * format, ... );
void error( char * format, ... );
int query( char * format, ... );

#ifdef __OS2__
static MRESULT EXPENTRY winproc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 );
#endif
#ifdef __WIN32__
static BOOL CALLBACK winproc( HWND hwnd, UINT msg, WPARAM mp1, LPARAM mp2 );
#define DosBeep Beep
#endif

static int hexdigit( int asciicode );
static int gethex( char * src, unsigned char * dst, int maxdstlen );
static int output( char * buf, int len );
static int outputs( char * buf );
static int printcal( void );
static void printday( int d );
static int  printweek( int d1 );
static void printmonth( int year, int month );

static int leap( int j );
static int fstday( int j );
static int easter( int J );
static void filldays( int year );


#ifdef __OS2__
void main( void )
/********************************************************************/
{
HAB hab;
HMQ hmq;

   hab = WinInitialize( 0 );
   hmq = WinCreateMsgQueue( hab, 0 );

   WinDlgBox( HWND_DESKTOP, HWND_DESKTOP, winproc, NULLHANDLE, 1, 0 );

   WinDestroyMsgQueue( hmq );
   WinTerminate( hab );
   exit( 0 );
}

static MRESULT EXPENTRY winproc( HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2 )
/********************************************************************/
{
int i;
char buf[1000];

   switch( msg )
   {
   case WM_INITDLG:
      /* Feldlaengen der Eingabefelder: */
      WinSendDlgItemMsg( hwnd, YEAR, EM_SETTEXTLIMIT, MPFROM2SHORT(4,0), NULL );
      WinSendDlgItemMsg( hwnd, OUTNAME, EM_SETTEXTLIMIT, MPFROM2SHORT(sizeof(outname)-1,0), NULL );
      WinSendDlgItemMsg( hwnd, PROLOG, EM_SETTEXTLIMIT, MPFROM2SHORT(3*sizeof(prolog),0), NULL );
      WinSendDlgItemMsg( hwnd, NORMAL, EM_SETTEXTLIMIT, MPFROM2SHORT(3*sizeof(normal),0), NULL );
      WinSendDlgItemMsg( hwnd, HILITE, EM_SETTEXTLIMIT, MPFROM2SHORT(3*sizeof(hilite),0), NULL );

      /* Anfangswerte einstellen: */
      WinSetDlgItemText( hwnd, YEAR, DEFAULTYEAR );
      for( i = 1; i <= 12; i++ )
         WinCheckButton( hwnd, MONTH+i, TRUE );
      WinSetDlgItemText( hwnd, OUTNAME, LPTPORT );
      WinSetDlgItemText( hwnd, PROLOG, LPT_PROLOG );
      WinSetDlgItemText( hwnd, NORMAL, LPT_NORMAL );
      WinSetDlgItemText( hwnd, HILITE, LPT_HILITE );
      
      break;

   case WM_COMMAND:
      switch( COMMANDMSG(&msg)->cmd )
      {
      case DID_OK:         /* Drucken */
         /* Werte abholen: */
         WinQueryDlgItemText( hwnd, YEAR, sizeof( buf ), buf );
         year = atoi( buf );

         for( i = 1; i <= 12; i++ )
            domonth[i] = WinQueryButtonCheckstate( hwnd, MONTH+i );

         WinQueryDlgItemText( hwnd, OUTNAME, sizeof( outname ), outname );

         WinQueryDlgItemText( hwnd, PROLOG, sizeof( buf ), buf );
         prolog_len = gethex( buf, prolog, sizeof(prolog) );
         WinQueryDlgItemText( hwnd, NORMAL, sizeof( buf ), buf );
         normal_len = gethex( buf, normal, sizeof(normal) );
         WinQueryDlgItemText( hwnd, HILITE, sizeof( buf ), buf );
         hilite_len = gethex( buf, hilite, sizeof(hilite) );

         printcal();       /* do it */
         return 0;

      case DID_CANCEL:     /* Ende */
         break;
      }
      break;
   }
   return WinDefDlgProc( hwnd, msg, mp1, mp2 );
}
#endif

#ifdef __WIN32__
#include <richedit.h>

int __pfnBkCheck;


int WINAPI WinMain( HINSTANCE hinst, HINSTANCE hprev, LPSTR cmdline, int showstate )
/********************************************************************/
{
   DialogBox( hinst, TEXT("KALENDER"), HWND_DESKTOP, winproc );
   return 0;
}

static BOOL CALLBACK winproc( HWND hwnd, UINT msg, WPARAM mp1, LPARAM mp2 )
/********************************************************************/
{
int i;
char buf[1000];

   switch( msg )
   {
   case WM_INITDIALOG:
      /* Feldlaengen der Eingabefelder: */
      /* passiert hier was ????? */
      SendDlgItemMessage( hwnd, YEAR,   EM_EXLIMITTEXT, 0, 4 );
      SendDlgItemMessage( hwnd, OUTNAME,EM_EXLIMITTEXT, 0, sizeof(outname)-1 );
      SendDlgItemMessage( hwnd, PROLOG, EM_EXLIMITTEXT, 0, 3*sizeof(prolog) );
      SendDlgItemMessage( hwnd, NORMAL, EM_EXLIMITTEXT, 0, 3*sizeof(normal) );
      SendDlgItemMessage( hwnd, HILITE, EM_EXLIMITTEXT, 0, 3*sizeof(hilite) );

      /* Anfangswerte einstellen: */
      SetDlgItemText( hwnd, YEAR, DEFAULTYEAR );
      for( i = 1; i <= 12; i++ )
         CheckDlgButton( hwnd, MONTH+i, TRUE );
      SetDlgItemText( hwnd, OUTNAME, LPTPORT );
      SetDlgItemText( hwnd, PROLOG, LPT_PROLOG );
      SetDlgItemText( hwnd, NORMAL, LPT_NORMAL );
      SetDlgItemText( hwnd, HILITE, LPT_HILITE );
      
      return TRUE;

   case WM_COMMAND:
      switch( LOWORD(mp1) )
      {
      case IDOK:         /* Drucken */
         /* Werte abholen: */
         GetDlgItemText( hwnd, YEAR, buf, sizeof( buf ) );
         year = atoi( buf );

         for( i = 1; i <= 12; i++ )
            domonth[i] = IsDlgButtonChecked( hwnd, MONTH+i ) == BST_CHECKED;

         GetDlgItemText( hwnd, OUTNAME, outname, sizeof( outname ) );

         GetDlgItemText( hwnd, PROLOG, buf, sizeof( buf ) );
         prolog_len = gethex( buf, prolog, sizeof(prolog) );
         GetDlgItemText( hwnd, NORMAL, buf, sizeof( buf ) );
         normal_len = gethex( buf, normal, sizeof(normal) );
         GetDlgItemText( hwnd, HILITE, buf, sizeof( buf ) );
         hilite_len = gethex( buf, hilite, sizeof(hilite) );

         printcal();       /* do it */
         return TRUE;

      case IDCANCEL:     /* Ende */
         EndDialog( hwnd, 0 );
         return TRUE;
      }
      break;
   }
   return FALSE;
}
#endif


static int hexdigit( int asciicode )
/********************************************************************/
{
   if( asciicode >= '0'  &&  asciicode <= '9' )
      return asciicode - '0';
   else if( asciicode >= 'A'  &&  asciicode <= 'F' )
      return asciicode - 'A' + 10;
   else if( asciicode >= 'a'  &&  asciicode <= 'f' )
      return asciicode - 'a' + 10;
   else
      return -300;  /* error */
}


static int gethex( char * src, unsigned char * dst, int maxdstlen )
/********************************************************************/
/* wandelt ASCII-Hexstring in src in Bytefolge in dst um.
   Laenge wird zurueckgegeben. Fehler -> 
   -1 zu viele Zeichen
   -2 Syntaxfehler
*/
{
int dstlen = 0;
int val;
   
   while( *src )
   {
      while( *src == ' ' )
         src++;

      if( *src )
      {
         if( dstlen > maxdstlen )
            return -1;              /* zu viele Zeichen */
         val = hexdigit( *src ) * 16 + hexdigit( *(src+1) );
         if( val < 0 )
            return -2;              /* falsches Zeichen */
         dst[dstlen++] = val;
         src += 2;   
      }
   }
   return dstlen;
}


int mprintf( char * format, ... )
/********************************************************************/
{
char buffer[1000];
va_list ap;
int num;

   va_start( ap, format );
   num = vsprintf( buffer, format, ap );
   va_end( ap );
   
#ifdef __OS2__   
   WinMessageBox( HWND_DESKTOP, HWND_DESKTOP, buffer, "mprintf", 0, MB_OK | MB_INFORMATION | MB_MOVEABLE );
#endif   
#ifdef __WIN32__   
   MessageBox( HWND_DESKTOP, buffer, "mprintf", MB_OK | MB_ICONINFORMATION );
#endif   

   return num;
}


void error( char * format, ... )
/********************************************************************/
{
char buffer[1000];
va_list ap;

   va_start( ap, format );
   vsprintf( buffer, format, ap );
   va_end( ap );
   
#ifdef __OS2__   
   WinMessageBox( HWND_DESKTOP, HWND_DESKTOP, buffer, NULL, 0, MB_ERROR );
#endif   
#ifdef __WIN32__   
   MessageBox( HWND_DESKTOP, buffer, NULL, MB_ICONERROR );
#endif   
}


int query( char * format, ... )
/********************************************************************/
/* returns: 1 for yes, 0 for no */
{
char buffer[1000];
va_list ap;

   va_start( ap, format );
   vsprintf( buffer, format, ap );
   va_end( ap );
   
#ifdef __OS2__   
   return WinMessageBox( HWND_DESKTOP, HWND_DESKTOP, buffer, "", 0, MB_YESNO | MB_QUERY ) == MBID_YES;
#endif   
#ifdef __WIN32__   
   return MessageBox( HWND_DESKTOP, buffer, "", MB_YESNO | MB_ICONQUESTION ) == IDYES;
#endif   
}


static int output( char * buf, int len )
/********************************************************************/
/* returns 1 if OK, 0 in case of an error */
{
int n;

   n = fwrite( buf, 1, len, outf );
   fflush( outf );
   if( n != len )
   {
      error( "Kann an '%s' nicht ausgeben !", outname );
      return 0;
   }
   return 1;
}


static int outputs( char * buf )
/********************************************************************/
{
   return output( buf, strlen(buf) );
}


static int printcal( void )
/********************************************************************/
/* returns 1 if OK, 0 in case of an error */
{
int i, n;

/* check values: */
   if( year < 1582  ||  year > 2399 )
   {
      error( "Es sind nur die Jahre von 1582 bis 2399 erlaubt !" );
      return 0;
   }

   n = 0;
   for( i = 1; i <= 12; i++ )
      n += !!domonth[i];         /* -> Anzahl der Monate */
   if( n == 0 )
   {
      error( "Es wurde kein Monat gewhlt " );
      return 0;
   }
   
   if( prolog_len < 0 )
   {
      error( "Der Vorspann %s !", prolog_len == -1 ? "ist zu lang" : "darf nur Hexadezimalzahlen enthalten" );
      return 0;
   }
   if( normal_len < 0 )
   {
      error( "Der 'normal'-String %s !", normal_len == -1 ? "ist zu lang" : "darf nur Hexadezimalzahlen enthalten" );
      return 0;
   }
   if( hilite_len < 0 )
   {
      error( "Der 'hervorgehoben'-String %s !", hilite_len == -1 ? "ist zu lang" : "darf nur Hexadezimalzahlen enthalten" );
      return 0;
   }

   outf = fopen( outname, "wb" );
   if( !outf )
   {
      error( "Kann '%s' nicht ffnen !", outname );
      return 0;
   }

   /* Jahreskalender -> day[] */
   filldays( year );

   if( query( "Drucke an '%s' ...", outname ) )
   {
      /* yes, drucken: */
      if( !output( prolog, prolog_len ) ) /* prolog ausgeben */
         return 0;

      /* nun die Monate: */
      for( i = 1; i <= 12; i++ )
         if( domonth[i] )
            printmonth( year, i );
   }
   else
   {
      /* no, nicht drucken */
      DosBeep( 110, 100 );
   }

   fclose( outf );
   return 1;
}


static void printday( int d )
/********************************************************************/
{
DAY * pd;
char buf[20];

   pd = day + d;

   sprintf( buf, "%s      %2d", dayname[pd->dayofweek], pd->date );

   if( pd->holiday )
   {
      output( hilite, hilite_len );
      outputs( buf );
      output( normal, normal_len );
   }
   else
      outputs( buf );

   if( pd->dayofweek == 6 )
      outputs( "" );
   else
      outputs( "" );
}


static int printweek( int d1 )
/********************************************************************/
/* returns the day after the week */
{
int month, d, i;

   month = day[d1].month;

   outputs( "" );

   /* eventuell ein paar leere Tage: */
   if( day[d1].date == 1 )
      for( i = 0; i < day[d1].dayofweek; i++ )
         outputs( "          " );

   /* Tage mit Text: */
   for( d = d1; d < (d1 + 7); d++ )
   {
      printday( d );
      if( day[d].dayofweek == 6 /*So*/ )
         break;
      if( d == last[month] )
      {
         for( i = day[d].dayofweek + 1; i <= 6 /*So*/; i++ )
         {
            if( i < 6 )
               outputs( "          " );
            else
               outputs( "          " );
         }
         break;
      }
   }
   outputs( "\r\n" );

   /* Rest der Wochenreihe: */
   for( i = 3; i <= WEEKLINES; i++ )
      outputs( "                                                                      \r\n" );

   return d + 1;
}


static void printmonth( int year, int month )
/********************************************************************/
{
int weeks, blanklines, i, w, d;
char buf[100];

   /* Anzahl der Wochen dieses Monats bestimmen: */
   weeks = day[last[month]].week - day[first[month]].week + 1;
   
   /* Anzahl der Leerzeilen am Seitenanfang: */
   blanklines = PAGELINES - (weeks * WEEKLINES + 2 );

   /* Leerzeilen ausgeben: */
   for( i = 1; i <= blanklines; i++ )
      output( "\r\n", 2 );

   /* Auf normal schalten, Monatsname ausgeben: */
   output( normal, normal_len );
   sprintf( buf, "%s %d\r\n", monname[month], year );
   outputs( buf );

   /* Jetzt kommt die erste Woche: */
   outputs( "ͻ\r\n" );

   d = printweek( first[month] );   

   /* zweite bis letzte Woche: */
   for( w = 2; w <= weeks; w++ )
   {   
      outputs( "Ķ\r\n" );

      d = printweek( d );
   }

   /* letzte Zeile: */
   outputs( "ͼ\014" );
}



static int leap( int j )
/********************************************************************/
/* Diese Funktion bestimmt, ob das uebergebene Jahr j ein Schaltjahr
   ist und gibt 1 zurueck, wenn es ein Schaltjahr ist, ansonsten 0. */
{
   return ((j % 4 == 0) && (j % 100 != 0)) || (j % 400 == 0);
}



static int fstday( int j )
/********************************************************************/
/* Diese Funktion bestimmt den Wochentag mit dem das uebergebene 
   Jahr beginnt. mo=0, di=1, ... so=6                               */
{
int f, i;
static int t[28] =
     {0,2,3,4,5,0,1,2,3,5,6,0,1,3,4,5,6,1,2,3,4,6,0,1,2,4,5,6};

   f = t[j % 28];
   i = 1700;

   do
   {
      if( j > i )
         f = f + 6;
      i = i + 100;
   } while( i <= 2300 );

   if( j > 2000 ) 
      f = f + 1;

   return f % 7;
}


static int easter( int J )
/********************************************************************/
/* Berechnung des Osterfestes, s. Enzyklopaedie Natur, 1960, S.106
   Die Funktion liefert das Datum der Ostersonntag im Maerz.
   (Wenn die Zahl groesser als 31 ist, dann ist es im April !)     
   ACHTUNG: nur von 1582..2399                                    */
{  
int q, a, b, c, D, M;
int z;

   if( J < 1582  ||  J > 2400 )
   {
      error( "Fuer %d kann das Osterfest nicht berechnet werden !", J );
      exit( 1 );
   }

   if( J <= 1699 )
   {
      D =  10;
      M = 202;
   }
   else if( J <= 1799 )
   {
      D =  11;
      M = 203;
   }
   else if( J <= 1899 )
   {
      D =  12;
      M = 203;
   }
   else if( J <= 2099 )
   {
      D =  13;
      M = 204;
   }
   else if( J <= 2199 )
   {
      D =  14;
      M = 204;
   }
   else if( J <= 2299 )
   {
      D =  15;
      M = 205;
   }
   else /* if( J <= 2399 ) */
   {
      D =  16;
      M = 206;
   }

   q = J / 4;
   a = J % 19;
   b = (M - 11 * a) % 30;
   if( (b>=28) && (M==202 || M==204 || M==205) )
      b--;
   c = (J + q + b - D) % 7;

   /* nach D.Herrmann, Algorithmen Arbeitsbuch, S.204, 
      muss verhindert werden, dass Ostern auf den 26.April
      faellt. 
      (U. Umstaenden muss auch der 25.April verhindert, da
      diese Umstaende unklar sind, fehlt das hier noch.) */
   
   z = 28 + b - c;

   if( z - 30 == 26 )
      z -= 7;
   
   return z;
}


static void filldays( int year )
/********************************************************************/
/* initialisiert day[] und first[] und last[] */
{
int i, nofdays, dow, wn, dat, mon, esun;
static int l[14] = {0,31,28,31,30,31,30,31,31,30,31,30,31,99};


   if( leap(year) )
   {                       /* Schaltjahr: */
      nofdays = 366;
      l[2] = 29;
   }
   else
   {                       /* normales Jahr: */
      nofdays = 365;
      l[2] = 28;
   }
   dow = fstday( year );   /* Wochentag des 1.Januar */
   wn  = 1;                /* Wochennummer           */
   mon = 1;                /* Monat                  */
   dat = 1;                /* Tag im Monat           */

   for( i = 0; i < nofdays+1; i++ )
   {
      /* erstmal alle Sonntage als Feiertage markieren: */
      day[i].holiday = dow == 6;

      /* nun die fixen Feiertage: */
      /* erstmal hier im Code, spaeter Dialog ... */
      if(   (dat== 1  &&  mon== 1)  /* Neujahr   */
         || (dat== 1  &&  mon== 5)  /* 1.Mai     */
         || (dat== 3  &&  mon==10)  /* Einheit   */
         || (dat==26  &&  mon==12)  /* Weihnacht */
         || (dat==25  &&  mon==12)  /* Weihnacht */
           ) day[i].holiday = 1;

      /* der ganze Rest ... */      
      day[i].dayofweek = dow;
      dow = (dow+1) % 7;

      day[i].week = wn;
      if( dow == 0 )    /* Wenn der naechste Tag Mo ist ...  */
         wn++;          /* dann beginnt dann eine neue Woche */

      /* ersten und letzten Tag jeden Monats merken: */
      if( dat == 1 )
         first[mon] = i;
      else if( dat == l[mon] )
         last[mon] = i;

      day[i].month = mon;
      day[i].date  = dat;
      dat++;
      if( dat > l[mon] )
      {
         mon++;
         dat = 1;
      }
   }

   /* Jetzt fehlen nur noch:
      Karfreitag           Ostersonntag -  2
      Ostermontag          Ostersonntag +  1
      Himmelfahrt          Ostersonntag + 39
      Pfingstmontag        Ostersonntag + 50
   */
   /* Index in day[] fuer Ostersonntag: 
      n-ter Maerz + Laenge von Jan und Feb, - 1, da ab 0 */
   esun = easter( year ) + l[1] + l[2] - 1;

   /* die o.g. Tage setzen: */
   day[esun- 2].holiday = 1;
   day[esun+ 1].holiday = 1;
   day[esun+39].holiday = 1;
   day[esun+50].holiday = 1;
}


